# -*- coding: utf-8 -*-
import os
from .text import camel_case2snake_case

_MODEL_IMPORT = """
from ..models import {model_name}
"""

_CLASS_AND_META = """
class {model_name}Factory(factory.django.DjangoModelFactory):
    class Meta:
        model = {model_name}
"""

_COMMON_FIELD = """
    {field_name} = {fake_expression}
"""

_FK_FIELD = """
    {field_name} = factory.SubFactory({related_factory_name})
"""

_M2M_FIELD = """
    @factory.post_generation
    def {field_name}(self, create, extracted, **_):
        if not create:
            return

        if extracted:
            for o in extracted:
                self.{field_name}.add(o)
"""


def build_triple_quote(s: str, format_dict: dict = None, clean_left: bool = True, clean_right: bool = False) -> str:
    if clean_left:
        s = s[1:]
    if clean_right:
        s = s[:-1]
    if format_dict is not None:
        s = s.format(**format_dict)
    return s


class FactoryTemplate:

    def __init__(self):
        self._import_set = set()
        self._local_import_lines = list()
        self._abs_import_lines = list()
        self._content_lines = list()

    def import_custom_lib(self, lib: str):
        if lib in self._import_set:
            return

        self._import_set.add(lib)
        if lib.startswith('from .'):
            self._local_import_lines.append(lib)
        else:
            self._abs_import_lines.append(lib)

    def _import_model(self, model_name: str):
        self._local_import_lines.append(
            build_triple_quote(_MODEL_IMPORT, {'model_name': model_name}, clean_left=False)
        )

    def import_related_factory(self, related_app_name: str, related_factory_name: str, current_app_name: str):
        if related_app_name == current_app_name:
            factory_filename = camel_case2snake_case(related_factory_name)
            s = f'from .{factory_filename} import {related_factory_name}'
        else:
            s = f'from {related_app_name}.factories import {related_factory_name}'

        self.import_custom_lib(s)

    def add_class_and_meta(self, model_name: str):
        self._content_lines.append(
            build_triple_quote(_CLASS_AND_META, {
                'model_name': model_name
            }, clean_left=False)
        )

    def add_common_field(self, field_name: str, fake_expression: str):
        self._content_lines.append(
            build_triple_quote(_COMMON_FIELD, {
                'field_name': field_name,
                'fake_expression': fake_expression,
            }, clean_right=True)
        )

    def add_fk_field(self, field_name: str, related_factory_name: str):
        self._content_lines.append(
            build_triple_quote(_FK_FIELD, {
                'field_name': field_name,
                'related_factory_name': related_factory_name,
            }, clean_right=True)
        )

    def add_m2m_field(self, field_name: str):
        self._content_lines.append(build_triple_quote(_M2M_FIELD, {
            'field_name': field_name
        }, clean_left=False))

    def save(self, folder: str, model_name: str):
        self._import_model(model_name)
        self._abs_import_lines.insert(0, 'import factory')

        filename = f'{camel_case2snake_case(model_name)}_factory.py'
        with open(os.path.join(folder, filename), 'w', encoding='utf-8') as fd:
            lines = list()
            lines.extend(self._abs_import_lines)
            lines.extend(self._local_import_lines)
            lines.extend(self._content_lines)
            content = '\n'.join(lines)
            if content[-1] != '\n':
                content += '\n'
            fd.write(content)



# -*- coding: utf-8 -*-
from typing import Dict, Type
from django.db import models

from .field_templates import BaseFieldTemplate,  IntegerFieldTemplate, BinaryFieldTemplate, BooleanFieldTemplate, \
    CharFieldTemplate, DateFieldTemplate, DateTimeFieldTemplate, DecimalFieldTemplate, DurationFieldTemplate, \
    EmailFieldTemplate, FileFieldTemplate, FilePathFieldTemplate, FloatFieldTemplate, GenericIPAddressFieldTemplate, \
    ImageFieldTemplate, JSONFieldTemplate, TextFieldTemplate, TimeFieldTemplate, \
    URLFieldTemplate, UUIDFieldTemplate, BigIntegerFieldTemplate, SmallIntegerFieldTemplate


class FieldTemplateTranslater:
    cached_mappings = dict()
    mappings: Dict[Type[models.Field], Type[BaseFieldTemplate]] = {
        models.BigIntegerField: BigIntegerFieldTemplate,
        models.BinaryField: BinaryFieldTemplate,
        models.BooleanField: BooleanFieldTemplate,
        models.CharField: CharFieldTemplate,
        models.DateField: DateFieldTemplate,
        models.DateTimeField: DateTimeFieldTemplate,
        models.DecimalField: DecimalFieldTemplate,
        models.DurationField: DurationFieldTemplate,
        models.EmailField: EmailFieldTemplate,
        models.FileField: FileFieldTemplate,
        models.FilePathField: FilePathFieldTemplate,
        models.FloatField: FloatFieldTemplate,
        models.GenericIPAddressField: GenericIPAddressFieldTemplate,
        models.ImageField: ImageFieldTemplate,
        models.IntegerField: IntegerFieldTemplate,
        models.JSONField: JSONFieldTemplate,
        models.PositiveIntegerField: IntegerFieldTemplate,
        models.PositiveBigIntegerField: BigIntegerFieldTemplate,
        models.PositiveSmallIntegerField: SmallIntegerFieldTemplate,
        models.SlugField: CharFieldTemplate,
        models.SmallIntegerField: IntegerFieldTemplate,
        models.TextField: TextFieldTemplate,
        models.TimeField: TimeFieldTemplate,
        models.URLField: URLFieldTemplate,
        models.UUIDField: UUIDFieldTemplate,
    }

    @classmethod
    def translate(cls, field_instance: models.Field) -> BaseFieldTemplate:
        field_template_cls = cls._get_field_template_cls(type(field_instance))
        return field_template_cls(field_instance)

    @classmethod
    def _get_field_template_cls(cls, field_cls) -> Type[BaseFieldTemplate]:
        cls_name = field_cls.__name__

        cache_result = cls.cached_mappings.get(cls_name, None)
        if cache_result is not None:
            return cache_result

        for f_cls in field_cls.__mro__:
            json_field_cls = cls.mappings.get(f_cls, None)
            if json_field_cls is not None:
                cls.cached_mappings[cls_name] = json_field_cls
                return json_field_cls

        raise TypeError(f"{cls_name} can not find corresponding field class.")



# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand

from dj3fb.core.factory_utils.factory_generator import FactoryGenerator


class Command(BaseCommand):

    def handle(self, *args, **options):
        FactoryGenerator().start()



# -*- coding: utf-8 -*-
import abc
from typing import Tuple, Optional
from django.db import models


class BaseFieldTemplate(abc.ABC):
    f_lib = 'from dj3fb.faker_utils import f'

    def __init__(self, field: models.Field):
        self.field = field

    @abc.abstractmethod
    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        raise NotImplementedError()


class SmallIntegerFieldTemplate(BaseFieldTemplate):
    delta = 0

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        if self.field.choices is not None:
            choices = [k for k, _ in self.field.choices]
            return None, f"factory.Iterator({str(choices)})"
        else:
            return None, f'factory.Sequence(lambda n: n + {self.delta})'


class IntegerFieldTemplate(SmallIntegerFieldTemplate):
    delta = 10000


class BigIntegerFieldTemplate(SmallIntegerFieldTemplate):
    delta = 100000000


class BinaryFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = "memoryview(f.binary(1024))"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class BooleanFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        return self.f_lib, "factory.Sequence(lambda _: f.boolean())"


class CharFieldTemplate(BaseFieldTemplate):
    field: models.CharField

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        if self.field.choices is not None:
            choices = [k for k, _ in self.field.choices]
            return None, f"factory.Iterator({str(choices)})"
        else:
            s = f"f.pystr(max_chars={self.field.max_length or 32})"
            return self.f_lib, f"factory.Sequence(lambda _: {s})"


class DateFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        return self.f_lib, "factory.Sequence(lambda _: f.date_this_century())"


class DateTimeFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        return self.f_lib, "factory.Sequence(lambda _: f.date_time_this_century())"


class DecimalFieldTemplate(BaseFieldTemplate):
    field: models.DecimalField

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        left_digits = None
        if self.field.max_digits is not None and self.field.decimal_places is not None:
            left_digits = self.field.max_digits - self.field.decimal_places

        s = f"f.pydecimal(left_digits={left_digits}, right_digits={self.field.decimal_places})"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class DurationFieldTemplate(BaseFieldTemplate):
    field: models.DurationField

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.time_delta()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class EmailFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.email()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class FilePathFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.file_path()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class FileFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        return None, f"factory.django.FileField()"


class FloatFieldTemplate(BaseFieldTemplate):
    field: models.FloatField

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.pyfloat()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class GenericIPAddressFieldTemplate(BaseFieldTemplate):
    field: models.GenericIPAddressField

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        if self.field.protocol == 'ipv6':
            s = "f.ipv4()"
        else:
            s = "f.ipv6()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class ImageFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        return None, f"factory.django.ImageField()"


class JSONFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.json()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class TextFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.text()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class TimeFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.time()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class URLFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.url()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"


class UUIDFieldTemplate(BaseFieldTemplate):

    def get_fake_expression(self) -> Tuple[Optional[str], str]:
        s = f"f.uuid4()"
        return self.f_lib, f"factory.Sequence(lambda _: {s})"



# -*- coding: utf-8 -*-
import os
from typing import Dict, List, Type, Optional
from collections import defaultdict
from django.conf import settings
from django.db import models
from django.apps import apps, AppConfig
from django.db.models.fields import AutoFieldMixin # noqa
from django.db.models.fields.related import RelatedField
from django.db.models.options import Options

from .factory_template import FactoryTemplate
from .field_template_translater import FieldTemplateTranslater
from .text import camel_case2snake_case

"""

FACTORY = {
    "existed_factories": {
        'django.contrib.auth.User': 'local_app.UserFactory'
    },
    "include_third_party_apps": [
        'third_party_app'
    ]
}

"""


class FactoryGenerator:

    def __init__(self):
        conf = getattr(settings, 'FACTORY', dict())
        self.include_third_party_apps: List[str] = conf.get('include_third_party_apps', list())
        self.existed_factories: Dict[str, str] = conf.get('existed_factories', dict())

        self._app_label2factories = defaultdict(list)
        self._app_label2app_name = dict()

    def start(self):
        target_apps: List[AppConfig] = self._get_target_apps()
        for app in target_apps:
            model_list = list(app.get_models())
            if len(model_list) == 0:
                continue

            factories_folder = self._ensure_factories_folder(app)
            for model in model_list:
                if self._check_existed_factory(model) is None:
                    self._build_model_factory(factories_folder, app, model)
            self._ensure_factories_package(factories_folder, app)

    def _check_existed_factory(self, model: Type[models.Model]) -> Optional[str]:
        app_label__model = f"{getattr(model, '_meta').app_config.name}.{model.__name__}"
        return self.existed_factories.get(app_label__model, None)

    def _build_model_factory(self, factories_folder: str, app: AppConfig, model: Type[models.Model]):
        def _get_related_factory_app(m: Type[models.Model]) -> AppConfig:
            meta: Options = getattr(m, '_meta')
            return meta.app_config

        template = FactoryTemplate()

        # 分析fields
        common_fields = dict()
        fk_fields = dict()
        m2m_fields = dict()
        opts: Options = getattr(model, '_meta')

        for f in opts.get_fields():
            if isinstance(f, RelatedField):
                if isinstance(f, models.ManyToManyField):
                    m2m_fields[f.name] = f
                else:
                    fk_fields[f.name] = f

                    related_model = f.related_model

                    related_app__factory = self._check_existed_factory(related_model)
                    if related_app__factory is not None:
                        app_name, factory_name = related_app__factory.rsplit('.', 1)
                        template.import_related_factory(
                            related_app_name=app_name,
                            related_factory_name=factory_name,
                            current_app_name=self._app_label2app_name[app.label]
                        )
                    else:
                        related_app = _get_related_factory_app(related_model)
                        template.import_related_factory(
                            related_app_name=self._app_label2app_name[related_app.label],
                            related_factory_name=f'{related_model.__name__}Factory',
                            current_app_name=self._app_label2app_name[app.label]
                        )
            elif isinstance(f, models.Field):
                # remove auto_field, auto_now fields
                if isinstance(f, AutoFieldMixin) or not f.editable:
                    continue

                common_fields[f.name] = f

        # class
        template.add_class_and_meta(model.__name__)

        for field_name, field_instance in common_fields.items():
            field_template = FieldTemplateTranslater.translate(field_instance)
            custom_lib, fake_expression = field_template.get_fake_expression()
            template.add_common_field(field_name=field_name, fake_expression=fake_expression)
            if custom_lib is not None:
                template.import_custom_lib(custom_lib)

        for field_name, field_instance in fk_fields.items():
            template.add_fk_field(
                field_name=field_name, related_factory_name=f'{field_instance.related_model.__name__}Factory'
            )

        for field_name, field_instance in m2m_fields.items():
            template.add_m2m_field(field_name)

        template.save(folder=factories_folder, model_name=model.__name__)
        self._app_label2factories[app.label].append(f'{model.__name__}Factory')

    def _ensure_factories_package(self, factories_folder: str, app: AppConfig):
        lines = list()

        factories = self._app_label2factories[app.label]
        for factory_name in factories:
            line = f'from .{camel_case2snake_case(factory_name)} import {factory_name}'
            lines.append(line)

        for app__factory in self.existed_factories.values():
            app_name, factory_name = app__factory.rsplit('.', 1)
            if app_name == app.name:
                line = f'from .{camel_case2snake_case(factory_name)} import {factory_name}'
                lines.append(line)

        with open(os.path.join(factories_folder, '__init__.py'), 'w', encoding='utf-8') as fd:
            content = '\n'.join(lines) + '\n'
            fd.write(content)

    @classmethod
    def _ensure_factories_folder(cls, app: AppConfig) -> str:
        factories_dir = os.path.join(app.path, 'factories')
        if not os.path.isdir(factories_dir):
            os.makedirs(factories_dir)
        return factories_dir

    def _get_target_apps(self) -> List[AppConfig]:
        target_apps = list()

        base_dir = str(settings.BASE_DIR)
        for app_name in settings.INSTALLED_APPS:
            app_label = str(app_name).split('.')[-1]
            self._app_label2app_name[app_label] = app_name
            app: AppConfig = apps.get_app_config(app_label)

            if str(app.path).startswith(base_dir):
                target_apps.append(app)
            else:
                # third-party
                if app_name in self.include_third_party_apps:
                    target_apps.append(app)

        return target_apps
